from __future__ import annotations

from loguru import logger

from aeon.core.liquid import LiquidLiteralBool
from aeon.core.liquid import LiquidVar
from aeon.core.substitutions import substitution_in_liquid
from aeon.core.substitutions import substitution_in_type
from aeon.core.terms import Var
from aeon.core.types import AbstractionType, TypeConstructor, TypeVar
from aeon.core.types import RefinedType
from aeon.core.types import Top
from aeon.core.types import Type
from aeon.core.types import TypePolymorphism
from aeon.core.types import t_unit
from aeon.typechecking.context import TypingContext
from aeon.utils.location import Location
from aeon.verification.vcs import Conjunction, UninterpretedFunctionDeclaration
from aeon.verification.vcs import Constraint
from aeon.verification.vcs import Implication
from aeon.verification.vcs import LiquidConstraint
from aeon.utils.name import Name, fresh_counter

ctrue = LiquidConstraint(LiquidLiteralBool(True))
cfalse = LiquidConstraint(LiquidLiteralBool(False))


def ensure_refined(t: Type) -> Type:
    """Ensures that the Base Types and TypeVars are refined. All other types remain the same."""
    match t:
        case RefinedType(_, _, _):
            return t
        case TypeVar(Name(name)):
            return RefinedType(Name(f"singleton_{name}", fresh_counter.fresh()), t, LiquidLiteralBool(True))
        case TypeConstructor(Name(name), _):
            return RefinedType(Name(f"singleton_{name}", fresh_counter.fresh()), t, LiquidLiteralBool(True))
        case _:
            return t


def is_first_order_function(at: AbstractionType):
    v: Type = at
    while isinstance(v, AbstractionType):
        match v.var_type:
            case AbstractionType(_, _, _):
                return False
            case Top() | RefinedType(_, _, _) | TypeVar(_) | TypeConstructor(_, _):
                pass
            case _:
                assert False
        v = v.type
    return True


def lower_constraint_type(ttype: Type) -> Type:
    match ttype:
        case TypeVar(name):
            return TypeConstructor(name)
        case Top():
            return t_unit
        case AbstractionType(name, b, r):
            return AbstractionType(name, lower_constraint_type(b), lower_constraint_type(r))
        case RefinedType(_, t, _):
            return lower_constraint_type(t)
        case TypeConstructor(name, args):
            if args:
                # Polymorphic types are represented by top
                return TypeConstructor(Name("Top", 0))
            else:
                return TypeConstructor(name)

        case _:
            assert False, f"Unsupport type in constraint {ttype} ({type(ttype)})"


def implication_constraint(name: Name, ty: Type, c: Constraint, loc: Location | None = None) -> Constraint:
    match ty:
        case Top() | TypeVar(_) | TypeConstructor(_, _):
            basety = lower_constraint_type(ty)
            assert isinstance(basety, TypeConstructor)
            return Implication(name, basety, LiquidLiteralBool(True), c, loc=loc)
        case RefinedType(tname, ttype, tref):
            ref_subs = substitution_in_liquid(tref, LiquidVar(name), tname)
            ltype = lower_constraint_type(ttype)
            assert isinstance(ltype, TypeConstructor) or isinstance(ltype, TypeVar)
            return Implication(name, ltype, ref_subs, c, loc=loc)
        case AbstractionType(_, _, _):
            # TODO Poly Refl: instead of true, reflect the implementation of the function?
            if is_first_order_function(ty):
                absty = lower_constraint_type(ty)
                assert isinstance(absty, AbstractionType)
                return UninterpretedFunctionDeclaration(name, absty, c)
            else:
                return c
        case TypePolymorphism(_, _, _):
            return c
        case _:
            assert False


def sub(ctx: TypingContext, t1: Type, t2: Type, loc: Location | None = None) -> Constraint:
    if t2 == Top():
        return ctrue
    match (ensure_refined(t1), ensure_refined(t2)):
        case RefinedType(n1, ty1, r1), RefinedType(n2, ty2, r2):
            if ty2 == Top():
                return ctrue
            if ty1 != ty2:
                return cfalse

            new_name: Name = Name(n1.name + n2.name, fresh_counter.fresh())

            # Performs substition on t2 to have the same name of t1
            r2_ = substitution_in_liquid(r2, LiquidVar(new_name), n2)
            r1_ = substitution_in_liquid(r1, LiquidVar(new_name), n1)
            lowert = lower_constraint_type(ty1)
            assert isinstance(lowert, TypeConstructor)
            rconstraint = Implication(new_name, lowert, r1_, LiquidConstraint(r2_, loc=loc), loc=loc)

            return rconstraint
        case TypePolymorphism(_, _, _), _:
            return ctrue
        case AbstractionType(a1, t1, rt1), AbstractionType(a2, t2, rt2):
            new_name_a: Name = Name(a1.name + a2.name, fresh_counter.fresh())

            c0 = sub(ctx, t2, t1, loc)
            rt1_ = substitution_in_type(rt1, Var(new_name_a), a1)
            rt2_ = substitution_in_type(rt2, Var(new_name_a), a2)
            c1 = sub(ctx, rt1_, rt2_, loc)
            return Conjunction(c0, implication_constraint(new_name_a, t2, c1, loc))
        case TypeConstructor(name1, args1), TypeConstructor(name2, args2):
            if name1 != name2:
                return cfalse
            if len(args1) != len(args2):
                return cfalse
            for it1, it2 in zip(args1, args2):
                if it1 != it2:  # TODO polytypes: subtyping here?
                    return cfalse
            return ctrue
        case _:
            logger.error(f"Failed subtyping by exhaustion: {t1} <: {t2}")
            return cfalse



import asyncio
from typing import List, Optional, AsyncIterable

from lsprotocol.types import (
    TEXT_DOCUMENT_CODE_ACTION,
    TEXT_DOCUMENT_COMPLETION,
    TEXT_DOCUMENT_DID_OPEN,
    TEXT_DOCUMENT_DID_CHANGE,
    WORKSPACE_DID_CHANGE_WATCHED_FILES,
    WORKSPACE_EXECUTE_COMMAND,
    ApplyWorkspaceEditParams,
    CodeAction,
    CodeActionKind,
    CodeActionOptions,
    CodeActionParams,
    Command,
    CompletionItem,
    CompletionOptions,
    CompletionParams,
    DidChangeTextDocumentParams,
    DidChangeWatchedFilesParams,
    DidOpenTextDocumentParams,
    Diagnostic,
    ExecuteCommandOptions,
    ExecuteCommandParams,
    MessageType,
    PublishDiagnosticsParams,
    Range,
    TextEdit,
    WorkspaceEdit,
)
from pygls.lsp.server import LanguageServer

from aeon.facade.driver import AeonDriver

SYNTHESIZERS = ["gp", "enumerative", "random_search", "synquid", "llm"]
SYNTHESIZE_COMMAND = "aeon.synthesize"


class AeonLanguageServer(LanguageServer):
    def __init__(self, aeon_driver: AeonDriver):
        super().__init__("aeon.lsp.server", "0.1.0")
        self.aeon_driver = aeon_driver
        self.debounce_delay = 0.3
        self._setup_handlers()

    def start(self, tcp_server):
        if not tcp_server:
            self.start_io()

        host, port = tcp_server.split(":") if ":" in tcp_server else ("localhost", tcp_server)

        print(f"Listening on {host}:{port}")
        self.start_tcp(host, int(port))

    async def _parse_and_send_diagnostics(self, uri) -> None:
        await asyncio.sleep(self.debounce_delay)
        diagnostics = []
        async for diag in self._get_diagnostics(uri):
            diagnostics.append(diag)
        self.text_document_publish_diagnostics(PublishDiagnosticsParams(uri=uri, diagnostics=diagnostics))

    async def _get_diagnostics(self, uri: str) -> AsyncIterable[Diagnostic]:
        from . import aeon_adapter

        ast = await aeon_adapter.parse(self, uri)
        for diag in ast.diagnostics:
            yield diag

    def _ranges_overlap(self, r1: Range, r2: Range) -> bool:
        """Return True if two LSP ranges overlap (touching at a boundary counts)."""
        if r1.end.line < r2.start.line or r2.end.line < r1.start.line:
            return False
        if r1.end.line == r2.start.line and r1.end.character < r2.start.character:
            return False
        if r2.end.line == r1.start.line and r2.end.character < r1.start.character:
            return False
        return True

    def _setup_handlers(self):
        @self.feature(TEXT_DOCUMENT_DID_OPEN)
        async def did_open(
            ls: AeonLanguageServer,
            params: DidOpenTextDocumentParams,
        ) -> None:
            await ls._parse_and_send_diagnostics(params.text_document.uri)

        @self.feature(TEXT_DOCUMENT_DID_CHANGE)
        async def did_change(
            ls: AeonLanguageServer,
            params: DidChangeTextDocumentParams,
        ) -> None:
            from . import aeon_adapter

            aeon_adapter.clear_cache(params.text_document.uri)
            await ls._parse_and_send_diagnostics(params.text_document.uri)

        @self.feature(WORKSPACE_DID_CHANGE_WATCHED_FILES)
        async def did_change_watched_file(
            ls: AeonLanguageServer,
            params: DidChangeWatchedFilesParams,
        ) -> None:
            from . import aeon_adapter

            for change in params.changes:
                aeon_adapter.clear_cache(change.uri)

        @self.feature(TEXT_DOCUMENT_COMPLETION, CompletionOptions(trigger_characters=["= "]))
        async def lsp_completion(
            ls: AeonLanguageServer,
            params: CompletionParams,
        ) -> Optional[List[CompletionItem]]:
            await asyncio.sleep(ls.debounce_delay)
            return []

        @self.feature(
            TEXT_DOCUMENT_CODE_ACTION,
            CodeActionOptions(code_action_kinds=[CodeActionKind.RefactorRewrite]),
        )
        async def code_action(
            ls: AeonLanguageServer,
            params: CodeActionParams,
        ) -> Optional[List[CodeAction]]:
            from . import aeon_adapter

            uri = params.text_document.uri
            cursor_range = params.range

            parse_result = await aeon_adapter.parse(ls, uri)
            if not parse_result.holes:
                return []

            actions = []
            for hole in parse_result.holes:
                if not ls._ranges_overlap(hole.range, cursor_range):
                    continue
                for synthesizer in SYNTHESIZERS:
                    action = CodeAction(
                        title=f"Synthesize ?{hole.name} with {synthesizer}",
                        kind=CodeActionKind.RefactorRewrite,
                        command=Command(
                            title=f"Synthesize ?{hole.name} with {synthesizer}",
                            command=SYNTHESIZE_COMMAND,
                            arguments=[uri, hole.name, synthesizer],
                        ),
                    )
                    actions.append(action)
            return actions

        @self.feature(
            WORKSPACE_EXECUTE_COMMAND,
            ExecuteCommandOptions(commands=[SYNTHESIZE_COMMAND]),
        )
        async def execute_command(
            ls: AeonLanguageServer,
            params: ExecuteCommandParams,
        ) -> None:
            if params.command != SYNTHESIZE_COMMAND:
                return

            args = params.arguments or []
            if len(args) < 3:
                ls.show_message("aeon.synthesize requires [uri, hole_name, synthesizer]", MessageType.Error)
                return

            uri, hole_name_str, synthesizer_name = args[0], args[1], args[2]

            loop = asyncio.get_event_loop()
            try:
                result = await loop.run_in_executor(
                    None,
                    lambda: _run_synthesis(ls.aeon_driver, ls, uri, hole_name_str, synthesizer_name),
                )
            except Exception as e:
                ls.show_message(f"Synthesis error: {e}", MessageType.Error)
                return

            if result is None:
                return

            synthesized_str, hole_range = result
            edit = TextEdit(range=hole_range, new_text=synthesized_str)
            workspace_edit = WorkspaceEdit(changes={uri: [edit]})
            ls.workspace_apply_edit(ApplyWorkspaceEditParams(edit=workspace_edit))
            ls.show_message(f"Synthesized ?{hole_name_str} = {synthesized_str}", MessageType.Info)


def _run_synthesis(driver: AeonDriver, ls: AeonLanguageServer, uri: str, hole_name_str: str, synthesizer_name: str):
    """Blocking synthesis function, meant to run in a thread executor."""
    from . import aeon_adapter
    from aeon.synthesis.entrypoint import synthesize_holes
    from aeon.synthesis.modules.synthesizerfactory import make_synthesizer
    from aeon.synthesis.uis.api import SilentSynthesisUI
    from aeon.sugar.lifting import lift
    from aeon.utils.pprint import pretty_print_sterm

    document = ls.workspace.get_text_document(uri)
    source = document.source

    try:
        errors = list(driver.parse(filename=uri, aeon_code=source))
    except Exception as e:
        ls.show_message(f"Cannot synthesize: parse failed ({e})", MessageType.Error)
        return None
    if errors:
        ls.show_message(f"Cannot synthesize: file has {len(errors)} error(s)", MessageType.Error)
        return None

    if not driver.has_synth():
        ls.show_message("No holes found in file", MessageType.Info)
        return None

    targets = [(fn, holes) for fn, holes in driver.incomplete_functions if any(h.name == hole_name_str for h in holes)]

    if not targets:
        ls.show_message(f"Hole ?{hole_name_str} not found or not synthesizable", MessageType.Warning)
        return None

    ls.show_message(f"Synthesizing ?{hole_name_str} with {synthesizer_name}...", MessageType.Info)

    try:
        synthesizer = make_synthesizer(synthesizer_name)
    except AssertionError:
        ls.show_message(f"Unknown synthesizer: {synthesizer_name}", MessageType.Error)
        return None

    ui = SilentSynthesisUI()
    try:
        mapping = synthesize_holes(
            driver.typing_ctx,
            driver.evaluation_ctx,
            driver.core,
            targets,
            driver.metadata,
            synthesizer,
            budget=5.0,
            ui=ui,
        )
    except Exception as e:
        ls.show_message(f"Synthesis failed: {e}", MessageType.Error)
        return None

    for hole_name, term in mapping.items():
        if hole_name.name == hole_name_str and term is not None:
            sterm = lift(term)
            synthesized_str = pretty_print_sterm(sterm)

            hole_positions = aeon_adapter.find_holes_in_source(source)
            hole_range = next(
                (h.range for h in hole_positions if h.name == hole_name_str),
                None,
            )
            if hole_range is None:
                ls.show_message(f"Could not locate ?{hole_name_str} in source", MessageType.Error)
                return None

            return synthesized_str, hole_range

    ls.show_message(f"No solution found for ?{hole_name_str}", MessageType.Warning)
    return None



from __future__ import annotations

import datetime
import sys

from loguru import logger


def levels_filter(levels):
    def filter(record):
        return record["level"].name in levels

    return filter


def setup_logger():
    for name, kwargs in [
        ("TYPECHECKER", {"no": 35, "color": "<magenta>", "icon": "🔍"}),
        ("SYNTH_TYPE", {"no": 36, "color": "<magenta>"}),
        ("CONSTRAINT", {"no": 37, "color": "<cyan>", "icon": "🔒"}),
        ("SYNTHESIZER", {"no": 38, "color": "<red>"}),
        ("TIME", {"no": 39, "color": "<green>"}),
    ]:
        try:
            logger.level(name, **kwargs)
        except ValueError:
            pass

    # Setup the logger
    logger.remove()
    # logger.add(sys.stderr, level="DEBUG")
    return logger


def export_log(logs: list, export_file: bool = False, logfile_name: str = ""):
    if export_file:
        logfile = f"logs/{logfile_name}_{datetime.datetime.now().strftime('%Y_%m_%d %H_%M_%S')}.log"
        return logger.add(logfile, filter=levels_filter(logs))
    else:
        return logger.add(sys.stderr, filter=levels_filter(logs))



from typing import Callable
from time import monotonic_ns

import numpy as np
from sklearn.tree import DecisionTreeRegressor

from aeon.core.terms import Application, If, Literal, Term, TypeApplication, Var
from aeon.core.types import Type, t_float
from aeon.decorators.api import Metadata
from aeon.synthesis.api import Synthesizer
from aeon.synthesis.uis.api import SynthesisUI
from aeon.typechecking.context import TypingContext, VariableBinder
from aeon.utils.name import Name


def get_elapsed_time(start_time) -> float:
    """The elapsed time since the start in seconds."""
    return (monotonic_ns() - start_time) * 0.000000001


def _get_function_arg_names(ctx: TypingContext, fun_name: Name) -> list[Name]:
    """Extract the function's argument names from the context.

    Arguments are the variables bound after the function itself in the context.
    """
    found_fun = False
    arg_names = []
    for entry in ctx.entries:
        if isinstance(entry, VariableBinder):
            if entry.name.name == fun_name.name:
                found_fun = True
                continue
            if found_fun:
                arg_names.append(entry.name)
    return arg_names


def _tree_to_term(tree: DecisionTreeRegressor, arg_names: list[Name]) -> Term:
    """Convert a fitted sklearn decision tree into nested If-then-else Terms."""
    tree_ = tree.tree_
    feature = tree_.feature
    threshold = tree_.threshold
    value = tree_.value

    def build(node_id: int) -> Term:
        if feature[node_id] == -2:
            # Leaf node: return the predicted value
            leaf_value = float(value[node_id].flatten()[0])
            return Literal(leaf_value, t_float)

        # Internal node: if arg_names[feature_idx] <= threshold then left else right
        feat_idx = feature[node_id]
        thresh = float(threshold[node_id])

        # <= is polymorphic (forall a:B, ...), so apply it to Float first
        le_typed = TypeApplication(Var(Name("<=", 0)), t_float)
        cond = Application(
            Application(le_typed, Var(arg_names[feat_idx])),
            Literal(thresh, t_float),
        )
        then_branch = build(tree_.children_left[node_id])
        else_branch = build(tree_.children_right[node_id])
        return If(cond, then_branch, else_branch)

    return build(0)


class DecisionTreeSynthesizer(Synthesizer):
    def __init__(self, max_depth: int | None = None):
        self.max_depth = max_depth

    def synthesize(
        self,
        ctx: TypingContext,
        type: Type,
        validate: Callable[[Term], bool],
        evaluate: Callable[[Term], list[float]],
        fun_name: Name,
        metadata: Metadata,
        budget: float = 60,
        ui: SynthesisUI = SynthesisUI(),
    ) -> Term:
        assert isinstance(ctx, TypingContext)
        assert isinstance(type, Type)

        start_time = monotonic_ns()

        # Extract training data from metadata (match by name string since IDs
        # may differ between parsing and binding phases)
        current_metadata = {}
        for key, val in metadata.items():
            if isinstance(val, dict) and key.name == fun_name.name:
                current_metadata = val
                break
        training_data: list[list[float]] = current_metadata.get("training_data", [])
        if not training_data:
            return None

        data = np.array(training_data)
        X = data[:, :-1]
        y = data[:, -1]

        # Extract argument names from the typing context
        arg_names = _get_function_arg_names(ctx, fun_name)
        if len(arg_names) != X.shape[1]:
            return None

        # Try increasing depths until budget runs out or validation passes
        best_term = None
        best_quality = None
        max_depth_limit = self.max_depth or X.shape[0]

        for depth in range(1, max_depth_limit + 1):
            if get_elapsed_time(start_time) > budget:
                break

            dt = DecisionTreeRegressor(max_depth=depth)
            dt.fit(X, y)

            candidate = _tree_to_term(dt, arg_names)
            elapsed = get_elapsed_time(start_time)

            if validate(candidate):
                try:
                    quality = evaluate(candidate)
                    is_best = best_quality is None or all(q <= bq for q, bq in zip(quality, best_quality))
                    ui.register(candidate, quality, elapsed, is_best)
                    if is_best:
                        best_term = candidate
                        best_quality = quality
                    # Perfect fit: no need to try deeper trees
                    if all(q == 0.0 for q in quality):
                        break
                except Exception:
                    ui.register(candidate, None, elapsed, False)
            else:
                ui.register(candidate, None, elapsed, False)

        return best_term



"""Meta-programming code for optimization-related decorators."""

import csv
import io
from typing import NamedTuple
from aeon.decorators.api import Metadata, metadata_update
from aeon.sugar.program import Definition, SApplication, STerm, SVar
from aeon.sugar.stypes import STypeConstructor
from aeon.sugar.ast_helpers import st_int, st_float
from aeon.utils.name import Name, fresh_counter

from aeon.sugar.program import SLiteral


def raise_decorator_error(name: str) -> None:
    raise Exception(f"Exception in decorator named {name}.")


class Goal(NamedTuple):
    minimize: bool
    length: int
    function: Name


def make_optimizer(
    args: list[STerm], fun: Definition, metadata: Metadata, typ: STypeConstructor, minimize: bool, length: int = 1
) -> tuple[Definition, list[Definition], Metadata]:
    """This decorator expects a single argument (the body of the definition).

    It does not modify the original definition, but appends a new
    definition to the program. This new definition has the name
    "_fitness_function", prefixed by the original definition's name
    """
    current_goals = metadata.get(fun.name, {}).get("goals", [])
    minimize_name = "minimize" if minimize else "maximize"
    function_name = Name(f"__internal__{minimize_name}_{fun.name}_{len(current_goals)}", fresh_counter.fresh())
    function = Definition(
        name=function_name,
        foralls=[],
        args=[],
        type=typ,
        body=args[0],
    )
    goal = Goal(minimize, length, function_name)

    metadata = metadata_update(
        metadata,
        fun,
        {
            "goals": current_goals + [goal],
        },
    )
    return fun, [function], metadata


def minimize_int(
    args: list[STerm],
    fun: Definition,
    metadata: Metadata,
) -> tuple[Definition, list[Definition], Metadata]:
    assert len(args) == 1, "minimize_int decorator expects a single argument"
    return make_optimizer(args, fun, metadata, st_int, minimize=True)


def maximize_int(
    args: list[STerm],
    fun: Definition,
    metadata: Metadata,
) -> tuple[Definition, list[Definition], Metadata]:
    assert len(args) == 1, "maximize_int decorator expects a single argument"
    return make_optimizer(args, fun, metadata, st_int, minimize=False)


def minimize_float(
    args: list[STerm],
    fun: Definition,
    metadata: Metadata,
) -> tuple[Definition, list[Definition], Metadata]:
    assert len(args) == 1, "minimize_float decorator expects a single argument"
    return make_optimizer(args, fun, metadata, st_float, minimize=True)


def maximize_float(
    args: list[STerm],
    fun: Definition,
    metadata: Metadata,
) -> tuple[Definition, list[Definition], Metadata]:
    assert len(args) == 1, "maximize_float decorator expects a single argument"
    return make_optimizer(args, fun, metadata, st_float, minimize=False)


def multi_minimize_float(
    args: list[STerm],
    fun: Definition,
    metadata: Metadata,
) -> tuple[Definition, list[Definition], Metadata]:
    """This decorator expects a single argument (the body of the definition)."""
    assert (
        len(
            args,
        )
        == 1
    ), "multi_minimize_float decorator expects a single argument"
    assert isinstance(args[1], SLiteral)
    assert isinstance(args[1].value, int), "multi_minimize_float decorator expects an integer argument"
    number_of_objectives = args[1].value
    return make_optimizer(
        args, fun, metadata, STypeConstructor(Name("List", 0)), minimize=True, length=number_of_objectives
    )


def hide(
    args: list[STerm],
    fun: Definition,
    metadata: Metadata,
) -> tuple[Definition, list[Definition], Metadata]:
    """This decorator expects more than zero arguments.

    It does not modify the original definition. It makes sure that no
    grammar nodes are generated from the var names passed as arguments.
    """
    assert len(args) != 0

    # TODO How can I verify if the function is in the context?
    def get_var_name(arg):
        if isinstance(arg, SVar):
            return arg.name
        else:
            raise_decorator_error("hide")

    # rethink this
    aux_dict = {"hide": [get_var_name(arg) for arg in args]}
    metadata = metadata_update(metadata, fun, aux_dict)

    return fun, [], metadata


def hide_types(
    args: list[STerm],
    fun: Definition,
    metadata: Metadata,
) -> tuple[Definition, list[Definition], Metadata]:
    """This decorator expects more than zero arguments.

    It does not modify the original definition. It makes sure that no
    grammar nodes are generated from the var names passed as arguments.
    """
    assert len(args) != 0

    # TODO How can I verify if the function is in the context?
    def get_var_name(arg):
        if isinstance(arg, SVar):
            return arg.name
        else:
            raise_decorator_error("hide_types")

    # rethink this
    aux_dict = {"hide_types": [get_var_name(arg) for arg in args]}
    metadata = metadata_update(metadata, fun, aux_dict)

    return fun, [], metadata


def error_fitness(
    args: list[STerm], fun: Definition, metadata: Metadata
) -> tuple[Definition, list[Definition], Metadata]:
    """This decorator expects one argument .

    It does not modify the original definition. It makes sure that
    the error fitness in case of any exception during the synthesis is the one defined in the argument
    """
    assert len(args) == 1

    aux_dict = {"error_fitness": args[0]}
    metadata = metadata_update(metadata, fun, aux_dict)

    return fun, [], metadata


def disable_control_flow(
    args: list[STerm], fun: Definition, metadata: Metadata
) -> tuple[Definition, list[Definition], Metadata]:
    """This decorator expects zero arguments .

    It does not modify the original definition. It makes sure that
    the control flow grammar nodes are allowed during synthesis
    """
    assert len(args) == 0

    aux_dict = {"disable_control_flow": True}
    metadata = metadata_update(metadata, fun, aux_dict)

    return fun, [], metadata


def allow_recursion(
    args: list[STerm],
    fun: Definition,
    metadata: Metadata,
) -> tuple[Definition, list[Definition], Metadata]:
    """This decorator expects a zero argument .

    It does not modify the original definition. It makes sure that
    recursion can be used during synthesis
    """
    assert len(args) == 0

    aux_dict = {"recursion": True}
    metadata = metadata_update(metadata, fun, aux_dict)

    return fun, [], metadata


def prompt(
    args: list[STerm],
    fun: Definition,
    metadata: Metadata,
) -> tuple[Definition, list[Definition], Metadata]:
    "Keeps track of the prompt that should be used in LLM-based synthesis"
    assert len(args) == 1
    assert isinstance(args[0], SLiteral) and isinstance(args[0].value, str)
    val = args[0].value
    metadata = metadata_update(metadata, fun, {"prompt": val})

    return fun, [], metadata


def _binop(op: str, left: STerm, right: STerm) -> STerm:
    """Build a binary operation AST node: (op left right)."""
    return SApplication(SApplication(SVar(Name(op, 0)), left), right)


def _squared(expr: STerm) -> STerm:
    """Build expr * expr (squared error)."""
    return _binop("*", expr, expr)


def _parse_csv_rows(text: str) -> list[list[float]]:
    """Parse CSV text into a list of rows of floats."""
    # Handle escaped newlines from string literals
    text = text.replace("\\n", "\n")
    reader = csv.reader(io.StringIO(text.strip()))
    rows = []
    for row in reader:
        if not row or all(cell.strip() == "" for cell in row):
            continue
        rows.append([float(cell.strip()) for cell in row])
    return rows


def _build_csv_fitness_body(rows: list[list[float]], fun_name: Name) -> STerm:
    """Build the fitness body: sum of abs(f(x1,...,xn) - expected) for each row."""
    assert len(rows) > 0, "CSV data must contain at least one row"
    n_cols = len(rows[0])
    assert n_cols >= 2, "CSV data must have at least 2 columns (inputs + expected output)"
    for i, row in enumerate(rows):
        assert len(row) == n_cols, f"CSV row {i} has {len(row)} columns, expected {n_cols}"

    error_terms = []
    for row in rows:
        inputs = row[:-1]
        expected = row[-1]

        # Build f(x1)(x2)...(xn)
        call: STerm = SVar(fun_name)
        for val in inputs:
            call = SApplication(call, SLiteral(val, st_float))

        # (call - expected)^2
        diff = _binop("-", call, SLiteral(expected, st_float))
        error_terms.append(_squared(diff))

    # Sum all error terms
    total = error_terms[0]
    for term in error_terms[1:]:
        total = _binop("+", total, term)

    return total


def csv_data(
    args: list[STerm],
    fun: Definition,
    metadata: Metadata,
) -> tuple[Definition, list[Definition], Metadata]:
    """Decorator that accepts inline CSV data as a string.

    Each row is a data point where the last column is the expected output
    and preceding columns are function arguments. Minimizes sum of absolute errors.

    Usage: @csv_data("1.0,2.0,3.0\\n4.0,5.0,12.0")
    """
    assert len(args) == 1, "csv_data decorator expects a single string argument"
    assert isinstance(args[0], SLiteral) and isinstance(args[0].value, str), (
        "csv_data decorator expects a string literal"
    )

    rows = _parse_csv_rows(args[0].value)
    body = _build_csv_fitness_body(rows, fun.name)
    current_data = metadata.get(fun.name, {}).get("training_data", [])
    metadata = metadata_update(metadata, fun, {"training_data": current_data + rows})
    return make_optimizer([body], fun, metadata, st_float, minimize=True)


def csv_file(
    args: list[STerm],
    fun: Definition,
    metadata: Metadata,
) -> tuple[Definition, list[Definition], Metadata]:
    """Decorator that accepts a CSV filename.

    Reads the file and behaves like csv_data: each row is a data point where
    the last column is the expected output and preceding columns are function arguments.
    Minimizes sum of absolute errors.

    Usage: @csv_file("data.csv")
    """
    assert len(args) == 1, "csv_file decorator expects a single string argument"
    assert isinstance(args[0], SLiteral) and isinstance(args[0].value, str), (
        "csv_file decorator expects a string literal"
    )

    filename = args[0].value
    with open(filename) as f:
        text = f.read()

    rows = _parse_csv_rows(text)
    body = _build_csv_fitness_body(rows, fun.name)
    current_data = metadata.get(fun.name, {}).get("training_data", [])
    metadata = metadata_update(metadata, fun, {"training_data": current_data + rows})
    return make_optimizer([body], fun, metadata, st_float, minimize=True)


def _extract_training_point(expr: STerm, fun_name: Name) -> list[float] | None:
    """Try to extract a training data point from a minimize expression.

    Looks for pattern: (fun_name(lit1)(lit2)...(litN)) - expected_lit
    Returns [lit1, ..., litN, expected_lit] or None if pattern doesn't match.
    """
    # Check for (- lhs rhs) pattern
    if not isinstance(expr, SApplication):
        return None
    if not isinstance(expr.fun, SApplication):
        return None
    minus_op = expr.fun.fun
    if not isinstance(minus_op, SVar) or minus_op.name.name != "-":
        return None

    lhs = expr.fun.arg  # f(x1)(x2)...(xn)
    rhs = expr.arg  # expected

    # Extract expected value
    if not isinstance(rhs, SLiteral) or not isinstance(rhs.value, (int, float)):
        return None
    expected = float(rhs.value)

    # Extract function call arguments (built right-to-left via currying)
    args: list[float] = []
    current = lhs
    while isinstance(current, SApplication):
        arg = current.arg
        if not isinstance(arg, SLiteral) or not isinstance(arg.value, (int, float)):
            return None
        args.append(float(arg.value))
        current = current.fun

    # Verify it's calling the right function
    if not isinstance(current, SVar) or current.name.name != fun_name.name:
        return None

    args.reverse()
    return args + [expected]


def _store_training_point(expr: STerm, fun: Definition, metadata: Metadata) -> Metadata:
    """If the expression matches f(args) - expected, store as training data."""
    point = _extract_training_point(expr, fun.name)
    if point is not None:
        current_data = metadata.get(fun.name, {}).get("training_data", [])
        metadata = metadata_update(metadata, fun, {"training_data": current_data + [point]})
    return metadata


def minimize(
    args: list[STerm],
    fun: Definition,
    metadata: Metadata,
) -> tuple[Definition, list[Definition], Metadata]:
    """Minimize decorator that also extracts training data when possible.

    Usage: @minimize(f(1.0, 2.0) - 3.0)

    Creates a fitness goal (like minimize_float) and, if the expression matches
    the pattern f(literal_args) - expected_literal, also stores a training data
    point for the decision tree synthesizer.
    """
    assert len(args) == 1, "minimize decorator expects a single argument"
    metadata = _store_training_point(args[0], fun, metadata)
    return make_optimizer(args, fun, metadata, st_float, minimize=True)


def maximize(
    args: list[STerm],
    fun: Definition,
    metadata: Metadata,
) -> tuple[Definition, list[Definition], Metadata]:
    """Maximize decorator converted to minimize, also extracts training data.

    Usage: @maximize(3.0 - f(1.0, 2.0))

    Internally converts to minimizing the negated expression (0 - expr).
    If the original expression matches the pattern f(literal_args) - expected_literal,
    stores a training data point for the decision tree synthesizer.
    """
    assert len(args) == 1, "maximize decorator expects a single argument"
    # Extract training data from the original expression
    metadata = _store_training_point(args[0], fun, metadata)
    # Convert maximize(expr) to minimize(0 - expr)
    negated = _binop("-", SLiteral(0.0, st_float), args[0])
    return make_optimizer([negated], fun, metadata, st_float, minimize=True)
